import db
from math import tanh

def dtanh(y):
    return 1.0-y*y

class Net:
    def __init__(self,dbname='nn.db'):
        self.db=db.Database(dbname)

    def getstrength(self,fromid,toid,layer):
        if layer==0: table="inputhidden"
        else: table="hiddenoutput"
        res=self.db.con.execute("select strength from %s where fromid=%d and toid=%d" % (table,fromid,toid)).fetchone()
        if res==None:
            if layer==0: return -0.2
            if layer==1: return 0
        return res[0]

    def setstrength(self,fromid,toid,layer,strength):
        if layer==0: table="inputhidden"
        else: table="hiddenoutput"
        res=self.db.con.execute("select rowid from %s where fromid=%d and toid=%d" % (table,fromid,toid)).fetchone()
        if res==None:
            self.db.con.execute("insert into %s (fromid,toid,strength) values (%d,%d,%f)" % (table,fromid,toid,strength))
        else:
            rowid=res[0]
            self.db.con.execute("update %s set strength=%f where rowid=%d" % (table,strength,rowid))

    def generatehiddennode(self,inputids,outputids):
        if len(inputids)>3: return None
        # Check if we already created a node for this set of words
        createkey='_'.join(sorted([str(wi) for wi in inputids]))
        res=self.db.con.execute("select rowid from hiddennode where create_key='%s'" % createkey).fetchone()
        # If not, create it
        if res==None:
            cur=self.db.con.execute("insert into hiddennode (create_key) values ('%s')" % createkey)
            hiddenid=cur.lastrowid
            # Put in some default weights
            for inputid in inputids:
                self.setstrength(inputid,hiddenid,0,1.0/len(inputids))
            for outputid in outputids:
                self.setstrength(hiddenid,outputid,1,0.1)
            self.db.con.commit()

    def getallhiddenids(self,inputids,outputids):
        l1={}
        for inputid in inputids:
            cur=self.db.con.execute("select toid from inputhidden where fromid=%d" % inputid)
            for row in cur: l1[row[0]]=1
        for outputid in outputids:
            cur=self.db.con.execute("select fromid from hiddenoutput where toid=%d" % outputid)
            for row in cur: l1[row[0]]=1
        return l1.keys()

    def setupnetwork(self,inputids,outputids):
        # Value lists
        self.inputids=inputids
        self.hiddenids=self.getallhiddenids(inputids, outputids)
        self.outputids=outputids

        # Node outputs
        self.ai=[1.0]*len(self.inputids)
        self.ah=[1.0]*len(self.hiddenids)
        self.ao=[1.0]*len(self.outputids)

        # Create weights matrix
        self.wi=[[self.getstrength(inputid,hiddenid,0) for hiddenid in self.hiddenids] for inputid in self.inputids]
        self.wo=[[self.getstrength(hiddenid,outputid,1) for outputid in self.outputids] for hiddenid in self.hiddenids]

    def feedforward(self):
        for i in range(len(self.inputids)):
            self.ai[i]=1.0
        # Hidden activations
        for j in range(len(self.hiddenids)):
            sum=0.0
            for i in range(len(self.inputids)):
                sum=sum+self.ai[i]*self.wi[i][j]
            self.ah[j]=tanh(sum)
        # Output activations
        for k in range(len(self.outputids)):
            sum=0.0
            for j in range(len(self.hiddenids)):
                sum=sum+self.ah[j]*self.wo[j][k]
            self.ao[k]=tanh(sum)
        return self.ao[:]

    def getresult(self,inputids,outputids):
        self.setupnetwork(inputids,outputids)
        return self.feedforward()

    def backpropagate(self,targets,n=0.5):
        # Calculate errors for output
        output_deltas=[0.0]*len(self.outputids)
        for k in range(len(self.outputids)):
            error=targets[k]-self.ao[k]
            output_deltas[k]=dtanh(self.ao[k])*error
        # Calculate errors for hidden layer
        hidden_deltas=[0.0]*len(self.hiddenids)
        for j in range(len(self.hiddenids)):
            error=0.0
            for k in range(len(self.outputids)):
                error=error+output_deltas[k]*self.wo[j][k]
            hidden_deltas[j]=dtanh(self.ah[j])*error
        # Update output weights
        for j in range(len(self.hiddenids)):
            for k in range(len(self.outputids)):
                change=output_deltas[k]*self.ah[j]
                self.wo[j][k]=self.wo[j][k]+n*change

    def updatedatabase(self):
        # Set weights to database values
        for i in range(len(self.inputids)):
            for j in range(len(self.hiddenids)):
                self.setstrength(self.inputids[i],self.hiddenids[j],0,self.wi[i][j])
        for j in range(len(self.hiddenids)):
            for k in range(len(self.outputids)):
                self.setstrength(self.hiddenids[j],self.outputids[k],1,self.wo[j][k])
        self.db.con.commit()

    def trainquery(self,inputids,outputids,selectedoutput):
        # Generate a hidden node if necessary
        self.generatehiddennode(inputids,outputids)
        self.setupnetwork(inputids,outputids)
        self.feedforward()
        targets=[0.0]*len(outputids)
        targets[outputids.index(selectedoutput)]=1.0
        self.backpropagate(targets)
        self.updatedatabase()

